import asyncio
from typing import List

import httpx

from obelisk.asynchronous.producer import \
    Producer as AsyncProducer
from obelisk.strategies.retry import RetryStrategy, \
    NoRetryStrategy
from obelisk.types import IngestMode, TimestampPrecision, \
    ObeliskKind


class Producer:
    """
    Synchronous equivalient of :class:`~obelisk.asynchronous.producer.Producer`,
    to publish data to Obelisk.
    """

    loop: asyncio.AbstractEventLoop
    async_producer: AsyncProducer

    def __init__(self, client: str, secret: str,
                 retry_strategy: RetryStrategy = NoRetryStrategy(),
                 kind: ObeliskKind = ObeliskKind.CLASSIC):
        self.async_producer = AsyncProducer(client, secret, retry_strategy, kind)
        self.loop = asyncio.get_event_loop()

    def send(self, dataset: str, data: List[dict],
             precision: TimestampPrecision = TimestampPrecision.MILLISECONDS,
             mode: IngestMode = IngestMode.DEFAULT) -> httpx.Response:
        """
        Publishes data to Obelisk

        Parameters
        ----------
        dataset : str
            ID for the dataset to publish to
        data : List[dict]
            List of Obelisk-acceptable datapoints.
            Exact format varies between Classic or HFS,
            caller is responsible for formatting.
        precision : TimestampPrecision = TimestampPrecision.MILLISECONDS
            Precision used in the numeric timestamps contained in data.
            Ensure it matches to avoid weird errors.
        mode : IngestMode = IngestMode.DEFAULT
            See docs for :class:`~obelisk.types.IngestMode`.

        Raises
        ------

        ObeliskError
            When the resulting status code is not 204, an empty :exc:`~obelisk.exceptions.ObeliskError` is raised.
        """

        task = self.loop.create_task(
            self.async_producer.send(dataset, data, precision, mode))
        return self.loop.run_until_complete(task)



import pytest
from .consumer import Consumer

pytest_plugins = ('pytest_asyncio',)

@pytest.mark.asyncio
async def test_demo_igent():
    consumer = Consumer(client="67c716e616c11421cfe2faf6", secret="08dafe89-0389-45b4-9832-cc565fb8c2eb")
    result = await consumer.single_chunk(
        datasets=["612f6c39cbceda0ea9753d95"],
        metrics=["org.dyamand.types.common.Temperature::number"],
        from_timestamp=1740924034000,
        to_timestamp=1741100614258,
        limit=2
    )

    assert len(result.items) == 2



from enum import Enum
from typing import List, Any, Optional

from pydantic import BaseModel


class IngestMode(str, Enum):
    """
    Classic Obelisk accepts three ways of submitting new data.
    This integrates with the concept of Streams,
    which is a way to process datapoints as they come in.

    The default submission method is to publish both to long-term storage
    and stream.
    This IngestMode can be changed to change this default.

    Does not apply to HFS
    """

    DEFAULT = 'default'
    STREAM_ONLY = 'stream_only'
    STORE_ONLY = 'store_only'


class TimestampPrecision(str, Enum):
    """
    When ingesting data it is important to specify which precision provided UNIX timestamps are in.
    If a provided timestamp is in seconds,
    but interpreted by Obelisk as milliseconds, it would erroneously be somewhere in the past.
    """

    __choices__ = ('SECONDS', 'MILLISECONDS', 'MICROSECONDS')

    SECONDS = 'seconds'
    MILLISECONDS = 'milliseconds'
    MICROSECONDS = 'microseconds'


class Datapoint(BaseModel, extra='allow'):
    timestamp: int
    value: Any
    dataset: Optional[str] = None
    metric: Optional[str] = None
    source: Optional[str] = None
    userId: Optional[int] = None  # Only if HFS and no other name for field


class QueryResult(BaseModel):
    items: List[Datapoint]
    cursor: Optional[str] = None


class ObeliskKind(str, Enum):
    CLASSIC = 'classic'
    HFS = 'hfs'



"""
Obelisk-py is a client for the Obelisk data platform.
We support both "classic" Obelisk and HFS,
each with a synchronous and async API.
The PyPi package name is ``obelisk-py``, the Python module is called ``obelisk``.

Your starting point will be one of the Consumer or Producer instances in :mod:`~.sync` or :mod:`~.asynchronous` depending on your preferred API.

Each of the `sync` or `asynchronous` modules will contain two main classes, those being Producer and Consumer.
Those submit or fetch data respectively.

Error handling
--------------

Obelisk-py comes with robust retry logic to handle any errors that may come up.
Issues like timeouts, temporary server errors or even DNS issues are fairly common, and handling them properly is important.
Each Client accepts a retry strategy of type :class:`~.strategies.retry.RetryStrategy`.
Several predefined strategies are available in :mod:`.strategies.retry`.
"""



from typing import List

import httpx

from obelisk.asynchronous.client import Client
from obelisk.exceptions import ObeliskError
from obelisk.types import IngestMode, TimestampPrecision


class Producer(Client):
    """
    Allows publishing of data to Obelisk.
    """

    async def send(self, dataset: str, data: List[dict],
                   precision: TimestampPrecision = TimestampPrecision.MILLISECONDS,
                   mode: IngestMode = IngestMode.DEFAULT) -> httpx.Response:
        """
        Publishes data to Obelisk

        Parameters
        ----------
        dataset : str
            ID for the dataset to publish to
        data : List[dict]
            List of Obelisk-acceptable datapoints.
            Exact format varies between Classic or HFS,
            caller is responsible for formatting.
        precision : :class:`~obelisk.types.TimestampPrecision` = TimestampPrecision.MILLISECONDS
            Precision used in the numeric timestamps contained in data.
            Ensure it matches to avoid weird errors.
        mode : :class:`~obelisk.types.IngestMode` = IngestMode.DEFAULT
            See docs for :class:`~obelisk.types.IngestMode`.

        Raises
        ------

        ObeliskError
            When the resulting status code is not 204, an empty :exc:`~obelisk.exceptions.ObeliskError` is raised.
        """

        params = {
            'datasetId': dataset,
            'timestampPrecision': precision.value,
            'mode': mode.value
        }

        response = await self.http_post(f'{self._ingest_url}/{dataset}', data=data,
                                        params=params)
        if response.status_code != 204:
            msg = f'An error occured during data ingest. Status {response.status_code}, message: {response.text}'
            self.log.warning(msg)
            raise ObeliskError(msg)
        return response



"""
This module contains the asynchronous API to Obelisk-py.
These methods all return a :any:`Coroutine`.

Relevant entrance points are :class:`.producer.Producer` to publish data to Obelisk,
or :class:`.consumer.Consumer` to consume data.
"""



import asyncio
from datetime import datetime, timedelta
from typing import List, Literal, Generator, Optional
from math import floor

from obelisk.asynchronous.consumer import \
    Consumer as AsyncConsumer
from obelisk.strategies.retry import RetryStrategy, \
    NoRetryStrategy
from obelisk.types import QueryResult, Datapoint, ObeliskKind


class Consumer:
    """
    Component that contains all the logic to consume data from
    the Obelisk API (e.g. historical data, sse).

    Wraps :class:`~obelisk.asynchronous.consumer.Consumer`.

    Obelisk API Documentation:
    https://obelisk.docs.apiary.io/
    """

    loop: asyncio.AbstractEventLoop
    """Event loop used to run interal async operations"""
    async_consumer: AsyncConsumer
    """The actual implementation this synchronous wrapper refers to"""

    def __init__(self, client: str, secret: str,
                 retry_strategy: RetryStrategy = NoRetryStrategy(),
                 kind: ObeliskKind = ObeliskKind.CLASSIC):
        self.async_consumer = AsyncConsumer(client, secret, retry_strategy, kind)
        self.loop = asyncio.get_event_loop()

    def single_chunk(self, datasets: List[str], metrics: Optional[List[str]] = None,
                     fields: Optional[dict] = None,
                     from_timestamp: Optional[int] = None, to_timestamp: Optional[int] = None,
                     order_by: Optional[dict] = None,
                     filter_: Optional[dict] = None,
                     limit: Optional[int] = None, limit_by: Optional[dict] = None,
                     cursor: Optional[str] = None) -> QueryResult:
        """
        Queries one chunk of events from Obelisk for given parameters,
        does not handle paging over Cursors.

        Parameters
        ----------

        datasets : List[str]
            List of Dataset IDs.
        metrics : Optional[List[str]] = None
            List of Metric IDs or wildcards (e.g. ``*::number``), defaults to all metrics.
        fields : Optional[List[str]] = None
            List of fields to return in the result set.
            Defaults to `[metric, source, value]`
        from_timestamp : Optional[int] = None
            Limit output to events after (and including)
            this UTC millisecond timestamp, if present.
        to_timestamp : Optional[int] = None
            Limit output to events before (and excluding)
            this UTC millisecond timestamp, if present.
        order_by : Optional[dict] = None
            Specifies the ordering of the output,
            defaults to ascending by timestamp.
            See Obelisk docs for format. Caller is responsible for validity.
        filter_ : Optional[dict] = None
            Limit output to events matching the specified Filter expression.
            See Obelisk docs, caller is responsible for validity.
        limit : Optional[int] = None
            Limit output to a maximum number of events.
            Also determines the page size.
            Default is server-determined, usually 2500.
        limit_by : Optional[dict] = None
            Limit the combination of a specific set of Index fields
            to a specified maximum number.
        cursor : Optional[str] = None
            Specifies the next cursor,
            used when paging through large result sets.
        """

        self.async_consumer.log.info("Starting task")
        task = self.loop.create_task(
            self.async_consumer.single_chunk(datasets, metrics, fields, from_timestamp,
                                             to_timestamp, order_by, filter_,
                                             limit, limit_by, cursor))
        self.async_consumer.log.info("Blocking...")
        return self.loop.run_until_complete(task)


    def query(self, datasets: List[str], metrics: Optional[List[str]] = None,
            fields: Optional[dict] = None,
            from_timestamp: Optional[int] = None,
            to_timestamp: Optional[int] = None,
            order_by: Optional[dict] = None,
            filter_: Optional[dict] = None,
            limit: Optional[int] = None,
            limit_by: Optional[dict] = None) -> List[Datapoint]:
        """
        Queries data from obelisk,
        automatically iterating when a cursor is returned.

        Parameters
        ----------

        datasets  : List[str]
            List of Dataset IDs.
        metrics : Optional[List[str]] = None
            List of Metric IDs or wildcards (e.g. `*::number`), defaults to all metrics.
        fields : Optional[List[str]] = None
            List of fields to return in the result set.
            Defaults to `[metric, source, value]`
        from_timestamp : Optional[int] = None
            Limit output to events after (and including)
            this UTC millisecond timestamp, if present.
        to_timestamp : Optional[int] = None
            Limit output to events before (and excluding)
            this UTC millisecond timestamp, if present.
        order_by : Optional[dict] = None
            Specifies the ordering of the output,
            defaults to ascending by timestamp.
            See Obelisk docs for format. Caller is responsible for validity.
        filter_ : Optional[dict] = None
            Limit output to events matching the specified Filter expression.
            See Obelisk docs, caller is responsible for validity.
        limit : Optional[int] = None
            Limit output to a maximum number of events.
            Also determines the page size.
            Default is server-determined, usually 2500.
        limit_by : Optional[dict] = None
            Limit the combination of a specific set of Index fields
            to a specified maximum number.
        """


        task = self.loop.create_task(
            self.async_consumer.query(datasets, metrics, fields, from_timestamp,
                                    to_timestamp, order_by, filter_, limit,
                                    limit_by))
        return self.loop.run_until_complete(task)


    def query_time_chunked(self, datasets: List[str], metrics: List[str],
                        from_time: datetime, to_time: datetime,
                        jump: timedelta, filter_: Optional[dict] = None,
                        direction: Literal['asc', 'desc'] = 'asc'
                        ) -> Generator[List[Datapoint], None, None]:
        """
        Fetches all data matching the provided filters,
        yielding one chunk at a time.
        One "chunk" may require several Obelisk calls to resolve cursors.
        By necessity iterates over time, no other ordering is supported.

        Parameters
        ----------

        datasets : List[str]
            Dataset IDs to query from
        metrics : List[str]
            IDs of metrics to query
        from_time : datetime.datetime
            Start time to fetch from
        to_time : datetime.datetime
            End time to fetch until.
        jump : datetime.timedelta
            Size of one yielded chunk
        filter_ : Optional[dict] = None
            Obelisk filter, caller is responsible for correct format
        direction : Literal['asc', 'desc'] = 'asc'
            Yield older data or newer data first, defaults to older first.
        """

        current_start = from_time
        while current_start < to_time:
            yield self.query(datasets=datasets, metrics=metrics,
                                from_timestamp=floor(current_start.timestamp() * 1000),
                                to_timestamp=floor((current_start + jump).timestamp() * 1000 - 1),
                                order_by={"field": ["timestamp"], "ordering": direction},
                                filter_=filter_)
            current_start += jump



from datetime import datetime, timedelta
import logging
import base64
from typing import Any, Optional

import httpx

from obelisk.exceptions import AuthenticationError
from obelisk.strategies.retry import RetryStrategy, \
    NoRetryStrategy
from obelisk.types import ObeliskKind


class Client:
    """
    Base class handling Obelisk auth and doing the core HTTP communication.
    Only exists in asynchronous variety, as it is not usually directly useful for user code.
    """

    _client: str = ""
    _secret: str = ""

    token: Optional[str] = None
    """Current authentication token"""
    token_expires: Optional[datetime] = None
    """Deadline after which token is no longer useable"""

    grace_period: timedelta = timedelta(seconds=10)
    """Controls how much before the expiration deadline a token will be refreshed."""
    retry_strategy: RetryStrategy
    kind: ObeliskKind

    log: logging.Logger

    _token_url = 'https://obelisk.ilabt.imec.be/api/v3/auth/token'
    _root_url = 'https://obelisk.ilabt.imec.be/api/v3'
    _metadata_url = 'https://obelisk.ilabt.imec.be/api/v3/catalog/graphql'
    _events_url = 'https://obelisk.ilabt.imec.be/api/v3/data/query/events'
    _ingest_url = 'https://obelisk.ilabt.imec.be/api/v3/data/ingest'
    _streams_url = 'https://obelisk.ilabt.imec.be/api/v3/data/streams'

    def __init__(self, client: str, secret: str,
                 retry_strategy: RetryStrategy = NoRetryStrategy(),
                 kind: ObeliskKind = ObeliskKind.CLASSIC) -> None:
        self._client = client
        self._secret = secret
        self.retry_strategy = retry_strategy
        self.kind = kind

        self.log = logging.getLogger('obelisk')

        if self.kind == ObeliskKind.HFS:
            self._token_url = 'https://obelisk-hfs.discover.ilabt.imec.be/auth/realms/obelisk-hfs/protocol/openid-connect/token'
            self._root_url = 'https://obelisk-hfs.discover.ilabt.imec.be'
            self._events_url = 'https://obelisk-hfs.discover.ilabt.imec.be/data/query/events'
            self._ingest_url = 'https://obelisk-hfs.discover.ilabt.imec.be/data/ingest'
        else:
            self._token_url = 'https://obelisk.ilabt.imec.be/api/v3/auth/token'
            self._root_url = 'https://obelisk.ilabt.imec.be/api/v3'
            self._metadata_url = 'https://obelisk.ilabt.imec.be/api/v3/catalog/graphql'
            self._events_url = 'https://obelisk.ilabt.imec.be/api/v3/data/query/events'
            self._ingest_url = 'https://obelisk.ilabt.imec.be/api/v3/data/ingest'
            self._streams_url = 'https://obelisk.ilabt.imec.be/api/v3/data/streams'

    async def _get_token(self):
        auth_string = str(base64.b64encode(
            f'{self._client}:{self._secret}'.encode('utf-8')), 'utf-8')
        headers = {
            'Authorization': f'Basic {auth_string}',
            'Content-Type': ('application/x-www-form-urlencoded'
                             if self.kind == ObeliskKind.HFS else 'application/json')
        }
        payload = {
            'grant_type': 'client_credentials'
        }

        async with httpx.AsyncClient() as client:
            response = None
            last_error = None
            retry = self.retry_strategy.make()
            while not response or await retry.should_retry():
                try:
                    request = await client.post(
                        self._token_url,
                        json=payload if self.kind == ObeliskKind.CLASSIC else None,
                        data=payload if self.kind == ObeliskKind.HFS else None,
                        headers=headers)

                    response = request.json()
                except Exception as e:
                    last_error = e
                    self.log.error(e)
                    continue

            if response is None and last_error is not None:
                raise last_error

            if request.status_code != 200:
                if 'error' in response:
                    self.log.warning(f"Could not authenticate, {response['error']}")
                    raise AuthenticationError

            self.token = response['access_token']
            self.token_expires = (datetime.now()
                                  + timedelta(seconds=response['expires_in']))

    async def _verify_token(self):
        if (self.token is None
                or self.token_expires < (datetime.now() - self.grace_period)):
            retry = self.retry_strategy.make()
            first = True
            while first or await retry.should_retry():
                first = False
                try:
                    await self._get_token()
                    return
                except:
                    continue

    async def http_post(self, url: str, data: Any = None,
                        params: Optional[dict] = None) -> httpx.Response:
        """
        Send an HTTP POST request to Obelisk,
        with proper auth.

        Possibly refreshes the authentication token and performs backoff as per `retry_strategy`.
        This method is not of stable latency because of these properties.

        No validation is performed on the input data,
        callers are responsible for formatting it in a method Obelisk understands.
        """

        await self._verify_token()

        headers = {
            'Authorization': f'Bearer {self.token}',
            'Content-Type': 'application/json'
        }
        if params is None:
            params = {}
        async with httpx.AsyncClient() as client:
            response = None
            retry = self.retry_strategy.make()
            last_error = None
            while not response or await retry.should_retry():
                if response is not None:
                    self.log.debug(f"Retrying, last response: {response.status_code}")

                try:
                    response = await client.post(url,
                                                 json=data,
                                                 params={k: v for k, v in params.items() if
                                                         v is not None},
                                                 headers=headers)

                    if response.status_code // 100 == 2:
                        return response
                except Exception as e:
                    self.log.error(e)
                    last_error = e
                    continue

            if not response and last_error:
                raise last_error
            return response
